package at.jku.ssw.set.tree.test;

import java.util.ArrayList;
import java.util.Arrays;

import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import at.jku.ssw.set.impl.InOrderTreeIterator;
import at.jku.students.set.tree.BinaryTreeSet;

public class BinaryTreeSetTest {

	protected BinaryTreeSet createTreeSet(int... objects) {
		BinaryTreeSet set = new BinaryTreeSet();
		for (int t : objects) {
			set.add(t);
		}
		return set;
	}

	private static ArrayList<Integer> subList(ArrayList<Integer> oldList, int startIndex, int lastIndex) {
		ArrayList<Integer> newList = new ArrayList<>();
		for (int i = startIndex; i < lastIndex; i++) {
			newList.add(oldList.get(i));
		}
		return newList;
	}

	public static ArrayList<ArrayList<Integer>> permutation(ArrayList<Integer> s) {
		ArrayList<ArrayList<Integer>> res = new ArrayList<ArrayList<Integer>>();
		if (s.size() == 1) {
			res.add(s);
		} else if (s.size() > 1) {
			int lastIndex = s.size() - 1;
			ArrayList<Integer> last = subList(s, lastIndex, s.size());
			ArrayList<Integer> rest = subList(s, 0, lastIndex);
			res = merge(permutation(rest), last);
		}
		return res;
	}

	public static ArrayList<ArrayList<Integer>> merge(ArrayList<ArrayList<Integer>> list, ArrayList<Integer> last) {
		ArrayList<ArrayList<Integer>> res = new ArrayList<>();
		for (ArrayList<Integer> s : list) {
			for (int i = 0; i <= s.size(); ++i) {
				ArrayList<Integer> ps = new ArrayList<>(s);
				for (int j = 0; j < last.size(); j++) {
					ps.add(i + j, last.get(j));
				}
				res.add(ps);
			}
		}
		return res;
	}

	@Rule
	public ExpectedException thrown = ExpectedException.none();

	@Test
	public void testAdd() {
		ArrayList<Integer> originalList = new ArrayList<>(Arrays.asList(1, 2, 3, 4, 5, 6, 7, 8, 9));
		for (ArrayList<Integer> l : permutation(originalList)) {
			BinaryTreeSet set = new BinaryTreeSet();
			for (Integer i : l) {
				set.add(i);
			}
			InOrderTreeIterator it = new InOrderTreeIterator(set.getRoot());
			ArrayList<Integer> integers = new ArrayList<>();
			while (it.hasNext()) {
				integers.add((Integer) it.next());
			}
			Assert.assertEquals(originalList, integers);
		}
	}

	@Test
	public void testAddRemove0() {
		ArrayList<Integer> originalList = new ArrayList<>(Arrays.asList(1, 2, 3, 4, 5, 6, 7, 8, 9));
		ArrayList<Integer> toRemove = new ArrayList<>(Arrays.asList(1, 2, 3, 4));
		ArrayList<Integer> toRemoveMissing = new ArrayList<>(Arrays.asList(42));
		ArrayList<Integer> removedList = new ArrayList<>(Arrays.asList(5, 6, 7, 8, 9));
		for (ArrayList<Integer> l : permutation(originalList)) {
			BinaryTreeSet set = new BinaryTreeSet();
			for (Integer i : l) {
				set.add(i);
			}
			for (Integer r : toRemove) {
				Assert.assertTrue(set.remove(r));
			}
			for (Integer r : toRemoveMissing) {
				Assert.assertFalse(set.remove(r));
			}
			InOrderTreeIterator it = new InOrderTreeIterator(set.getRoot());
			ArrayList<Integer> integers = new ArrayList<>();
			while (it.hasNext()) {
				integers.add((Integer) it.next());
			}
			Assert.assertEquals(removedList, integers);
		}
	}

	@Test
	public void testAddRemove1() {
		int[] insertions = { 4, 1, 2 };
		BinaryTreeSet set = new BinaryTreeSet();
		for (int value : insertions) {
			set.add(value);
		}
		Assert.assertTrue(set.remove(2));
	}

	@Test
	public void testEmptyHeight() {
		BinaryTreeSet set = createTreeSet();
		Assert.assertEquals(0, set.height());
	}

	@Test
	public void testHeight1() {
		BinaryTreeSet set = createTreeSet(1, 2, 3, 4, 5, 6);
		Assert.assertEquals(6, set.height());
	}

	@Test
	public void testHeight2() {
		BinaryTreeSet set = createTreeSet(1, 2, 3, 4, 5, 0);
		Assert.assertEquals(5, set.height());
	}

	@Test
	public void testEmptyWdith() {
		BinaryTreeSet set = createTreeSet();
		Assert.assertEquals(0, set.width());
	}

	@Test
	public void testWidth() {
		BinaryTreeSet set = createTreeSet(4, 2, 1, 3, 6, 5, 7, 0);
		Assert.assertEquals(4, set.width());
	}

	@Test
	public void testSize() {
		BinaryTreeSet set = createTreeSet(4, 2, 1, 3, 6, 5, 7, 0);
		Assert.assertEquals(8, set.size());
	}

	@Test
	public void testSizeEmpty() {
		BinaryTreeSet set = createTreeSet();
		Assert.assertEquals(0, set.size());
	}

	@Test
	public void testAddContains() {
		BinaryTreeSet set = createTreeSet(0);
		set.add(0);
		Assert.assertEquals(1, set.size());
		Assert.assertEquals(true, set.contains(0));
	}

	@Test
	public void testTreeString() {
		BinaryTreeSet set = new BinaryTreeSet();
		set.add(2);
		set.add(3);
		set.add(1);
		String s = set.treeString();

		StringBuilder sb = new StringBuilder();
		sb.append("=> 2").append(System.lineSeparator());
		sb.append("  => 1").append(System.lineSeparator());
		sb.append("  => 3").append(System.lineSeparator());
		Assert.assertEquals(sb.toString(), s);

	}

}
